// 
// Copyright (C) 2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA

/*---------------------------------------------------------------------------*\
 |                                                                           |
 |  COBOL Compiler Run Time Library -- Pictures                              |
 |    See each function's comments for more detail.                          |
 |  Independent of compressed PIC implementation:                            |
 |    tcob_picExpand() - Returns an expanded version of the PIC string       |
 |    tcob_picCompLength() - Numeric character count                         |
 |    tcob_picEditedCompLength() - Numeric-edited character count            |
 |    tcob_picEditedCompDecimals() - Decimal numeric-edited characters count |
 |  Dependent on compressed PIC implementation:                              |
 |    tcob_picElemVal() - Retrieve character value from a specified element  |
 |    tcob_picElemLen() - Retrieve character count from a specified element  |
 |    tcob_picReqLen() - Determine size to allocate for a new compressed PIC |
 |    tcob_picCreate() - Start a new compressed PIC string                   |
 |    tcob_picAppend() - Continue a compressed PIC string                    |
 |                                                                           |
\*---------------------------------------------------------------------------*/

#include "htcoblib.h"

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_picExpand                                |
 |  expand a picture returning a malloc'ed string                         |
 |  from: "XAYBZC" (where X, Y, Z are picture characters and              |
 |                  A, B, C are binary counts for each character)         |
 |    to: "XXXXYYYZZZZZ"                                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/

char * tcob_picExpand( struct fld_desc *f ) {
	char            *result, c;
	unsigned int    i, t, tot;

	for (i=0, tot=0; tcob_picElemVal(f->pic,i); i++)
		tot += tcob_picElemLen(f->pic,i);
	result = malloc(tot+1);
	for (i=0, tot=0; (c=tcob_picElemVal(f->pic,i)); i++) {
		t = tcob_picElemLen(f->pic,i);
		memset (&result[tot], c, t);
		tot += t;
	}
	result[tot] = '\0';
/*
#ifdef PICTURE_TESTING
	printf("tcob_picexpand: %s\n", result);
	free(result);
	return NULL;
#else
	return result;
#endif
*/
	return result;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_picCompLength                            |
 |  For numeric types, determine the total number of stored digits.       |
 |  Includes 9's.                                                         |
 |                                                                        |
\*------------------------------------------------------------------------*/

unsigned int tcob_picCompLength( struct fld_desc *f ) {
	unsigned int  len=0, i;
	char c;
	
	for (i=0; (c=tcob_picElemVal(f->pic,i)); i++) {
		if (c == '9')
			len += tcob_picElemLen(f->pic,i);
	}
	return len;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_picEditedCompLength                      |
 |  For the numeric-edited type, determine the total number of digits     |
 |  available.  Includes 9's, Z's,*'s, and multiples of +, -, and the     |
 |  currency sign.                                                        |
 |                                                                        |
\*------------------------------------------------------------------------*/

unsigned int tcob_picEditedCompLength( struct fld_desc *f ) {
	unsigned int  len=0, i;
	unsigned int bSeenPlus=0, bSeenMinus=0, bSeenCurrency=0;
	char c;
	extern char cCurrencySymbol;

	for (i=0; (c=tcob_picElemVal(f->pic,i)); i++) {
		unsigned int ilen = tcob_picElemLen(f->pic,i);
		if (c == '9' || c == 'Z' || c == '*' ||
		    (bSeenPlus && (c == '+')) ||
		    (bSeenMinus && (c == '-')) ||
		    (bSeenCurrency && (c == cCurrencySymbol)) )
			len += ilen;
		
		if (!bSeenPlus && (c == '+')) {
			bSeenPlus = 1;
			len += ilen - 1;
		} else if (!bSeenMinus && (c == '-')) {
			bSeenMinus = 1;
			len += ilen - 1;
		} else if (!bSeenCurrency && (c == cCurrencySymbol)) {
			bSeenCurrency = 1;
			len += ilen - 1;
		}
	}
	return len;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_picEditedCompDecimals                    |
 |  For the numeric-edited type, determine the total number of digits     |
 |  available after the decimal point.  Includes 9's, Z's,*'s, and        |
 |  multiples of +, -, and the currency sign.                             |
 |                                                                        |
\*------------------------------------------------------------------------*/

unsigned int tcob_picEditedCompDecimals( struct fld_desc *f ) {
	unsigned int  len=0, i;
	int bSeenDecimal=0, bSeenPlus=0, bSeenMinus=0, bSeenCurrency=0;
	extern int bDecimalComma;
	extern char cCurrencySymbol;
	char c, cDecimal = (bDecimalComma) ? ',' : '.';

	for (i=0; (c=tcob_picElemVal(f->pic,i)); i++) {
		unsigned int ilen = tcob_picElemLen(f->pic,i);
		if (bSeenDecimal &&  // Must find the decimal point first
		    (c == '9' || c == 'Z' || c == '*' ||
		    (bSeenPlus && (c == '+')) ||
		    (bSeenMinus && (c == '-')) ||
		    (bSeenCurrency && (c == cCurrencySymbol)) ) )
			len += ilen;
		if (!bSeenDecimal && (c == cDecimal)) {
			bSeenDecimal = 1;
		} else if (!bSeenPlus && (c == '+')) {
			bSeenPlus = 1;
			if (bSeenDecimal)
				len += ilen - 1;
		} else if (!bSeenMinus && (c == '-')) {
			bSeenMinus = 1;
			if (bSeenDecimal)
				len += ilen - 1;
		} else if (!bSeenCurrency && (c == cCurrencySymbol)) {
			bSeenCurrency = 1;
			if (bSeenDecimal)
				len += ilen - 1;
		}
	}
	return len;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                   tcob_picElemLen & tcob_picElemVal                    |
 |  Determine the length or value of element i in array p.                |
 |  Each element is composed of an 8-bit character and an 8-bit length.   |
 |                                                                        |
\*------------------------------------------------------------------------*/

unsigned int tcob_picElemLen(char *p, unsigned int i) {
	return ( (unsigned int)(((unsigned short *)p)[i] >> 8) );
}

char tcob_picElemVal(char *p, unsigned int i) {
	return ( ((unsigned short *)p)[i] & 0x0ff );
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                           tcob_picReqLen                               |
 |  Determine number of bytes to request for holding a compressed PIC     |
 |  string of i elements.                                                 |
 |                                                                        |
\*------------------------------------------------------------------------*/

unsigned int tcob_picReqLen(unsigned int i) {
	/* Currently using a 1-byte value and a 1-byte length for each
	 * element, plus a 1-byte NULL to indicate the end */
	return ((i*2)+1);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                           tcob_picCreate                               |
 |  Create a new PIC string and place in p.                               |
 |                                                                        |
\*------------------------------------------------------------------------*/

char *tcob_picCreate(char *p, unsigned int len, ...) {
	char pc;
	unsigned int i=0, ilen;
	va_list args;

	va_start(args, len);
	while ((pc=va_arg(args,int))) {
		for (ilen=va_arg(args, int); ilen>255; ilen-=255) {
			if (i+1>=len)
				break;
			p[i] = pc;
			p[i+1] = 255;
			ilen -= 255;
			i += 2;
		}
		if (i+1>=len)
			break;
		p[i] = pc;
		p[i+1] = ilen;
		i += 2;
	}
	if (i<len)
		p[i] = '\0';
	  else if (i>1)
		p[i-2] = '\0';
	  else if (len==1)
		p[0] = '\0';
	va_end(args);
	return p;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                           tcob_picAppend                               |
 |  Continue the PIC string found in p.                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/

char *tcob_picAppend(char *p, unsigned int len, ...) {
	char pc;
	unsigned int i, ilen;
	va_list args;
	
	for (i=0; p[i]; i+=2) ;

	va_start(args, len);
	while ((pc=va_arg(args,int))) {
		for (ilen=va_arg(args, int); ilen>255; ilen-=255) {
			if (i+1>=len)
				break;
			p[i] = pc;
			p[i+1] = 255;
			ilen -= 255;
			i += 2;
		}
		if (i+1>=len)
			break;
		p[i] = pc;
		p[i+1] = ilen;
		i += 2;
	}
	if (i<len)
		p[i] = '\0';
	  else if (i>1)
		p[i-2] = '\0';
	  else if (len==1)
		p[0] = '\0';
	va_end(args);
	return p;
}
